#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------
"""Provides routines to support the geomagnetic indices, Kp and Ap."""

import datetime as dt
import numpy as np
from packaging.version import Version
import pandas as pds

import pysat

import pysatSpaceWeather as pysat_sw
from pysatSpaceWeather.instruments.methods import gfz
from pysatSpaceWeather.instruments.methods import swpc


# --------------------------------------------------------------------------
# Instrument utilities

def acknowledgements(name, tag):
    """Define the acknowledgements for the geomagnetic data sets.

    Parameters
    ----------
    name : str
        Instrument name of space weather index, accepts 'kp' or 'ap'.
    tag : str
        Instrument tag.

    """

    ackn = {'kp': {'': 'Provided by GFZ German Research Centre for Geosciences',
                   'forecast': swpc.ackn, 'recent': swpc.ackn, 'def': gfz.ackn,
                   'now': gfz.ackn, 'prediction': swpc.ackn},
            'ap': {'forecast': swpc.ackn, 'recent': swpc.ackn,
                   'prediction': swpc.ackn, '45day': swpc.ackn,
                   'def': gfz.ackn, 'now': gfz.ackn},
            'cp': {'def': gfz.ackn, 'now': gfz.ackn}}

    return ackn[name][tag]


def references(name, tag):
    """Define the references for the geomagnetic data sets.

    Parameters
    ----------
    name : str
        Instrument name of space weather index, accepts kp or ap.
    tag : str
        Instrument tag.

    """

    gen_refs = "\n".join([''.join(["J. Bartels, The technique of scaling ",
                                   "indices K and Q of geomagnetic activity, ",
                                   "Ann. Intern. Geophys. Year 4, 215-226, ",
                                   "1957."]),
                          ''.join(["J. Bartels,The geomagnetic measures for ",
                                   "the time-variations of solar corpuscular ",
                                   "radiation, described for use in ",
                                   "correlation studies in other geophysical ",
                                   "fields, Ann. Intern. Geophys. Year 4, ",
                                   "227-236, 1957."]),
                          ''.join(["P.N. Mayaud, Derivation, Meaning and Use ",
                                   "of Geomagnetic Indices, Geophysical ",
                                   "Monograph 22, Am. Geophys. Union, ",
                                   "Washington D.C., 1980."]),
                          ''.join(["G.K. Rangarajan, Indices of magnetic ",
                                   "activity, in Geomagnetism, edited by I.A. ",
                                   "Jacobs, Academic, San Diego, 1989."]),
                          ''.join(["M. Menvielle and A. Berthelier, The ",
                                   "K-derived planetary indices: description ",
                                   "and availability, Rev. Geophys. 29, 3, ",
                                   "415-432, 1991."])])

    refs = {'kp': {'': gen_refs, 'forecast': gen_refs, 'recent': gen_refs,
                   'prediction': gen_refs, 'def': gfz.geoind_refs,
                   'now': gfz.geoind_refs},
            'ap': {'recent': gen_refs, 'forecast': gen_refs, '45day': gen_refs,
                   'prediction': gen_refs, 'def': gfz.geoind_refs,
                   'now': gfz.geoind_refs},
            'cp': {'def': gfz.geoind_refs, 'now': gfz.geoind_refs}}

    return refs[name][tag]


def initialize_kp_metadata(meta, data_key, fill_val=-1):
    """Initialize the Kp meta data using our knowledge of the index.

    Parameters
    ----------
    meta : pysat._meta.Meta
        Pysat Metadata
    data_key : str
        String denoting the data key
    fill_val : int or float
        File-specific fill value (default=-1)

    """

    meta[data_key] = {meta.labels.units: '',
                      meta.labels.name: data_key,
                      meta.labels.desc: "Planetary K-index",
                      meta.labels.min_val: 0,
                      meta.labels.max_val: 9,
                      meta.labels.fill_val: fill_val}

    return


def initialize_ap_metadata(meta, data_key, fill_val=-1):
    """Initialize the ap meta data using our knowledge of the index.

    Parameters
    ----------
    meta : pysat._meta.Meta
        Pysat Metadata
    data_key : str
        String denoting the data key
    fill_val : int or float
        File-specific fill value (default=-1)

    """

    meta[data_key] = {meta.labels.units: '',
                      meta.labels.name: data_key,
                      meta.labels.desc: "ap (equivalent range) index",
                      meta.labels.min_val: 0,
                      meta.labels.max_val: 400,
                      meta.labels.fill_val: fill_val}

    return


def initialize_bartel_metadata(meta, data_key, fill_val=-1):
    """Initialize the Bartel rotation meta data using our knowledge of the data.

    Parameters
    ----------
    meta : pysat._meta.Meta
        Pysat Metadata
    data_key : str
        String denoting the data key
    fill_val : int or float
        File-specific fill value (default=-1)

    """

    if data_key.find('solar_rotation_num') >= 0:
        units = ''
        bname = 'Bartels solar rotation number'
        desc = 'A sequence of 27-day intervals counted from February 8, 1832'
        max_val = np.inf
    elif data_key.find('day_within') >= 0:
        units = 'day'
        bname = 'Days within Bartels solar rotation'
        desc = 'Number of day within the Bartels solar rotation',
        max_val = 27
    else:
        raise ValueError('unknown data key: {:}'.format(data_key))

    meta[data_key] = {meta.labels.units: units,
                      meta.labels.name: bname,
                      meta.labels.desc: desc,
                      meta.labels.min_val: 1,
                      meta.labels.max_val: max_val,
                      meta.labels.fill_val: fill_val}

    return

# --------------------------------------------------------------------------
# Common custom functions


def convert_3hr_kp_to_ap(kp_inst, var_name='Kp'):
    """Calculate 3 hour ap from 3 hour Kp index.

    Parameters
    ----------
    kp_inst : pysat.Instrument
        Instrument containing Kp data
    var_name : str
        Variable name for the Kp data (default='Kp')

    Raises
    ------
    ValueError
        If `var_name` is not present in `kp_inst`.

    Note
    ----
    Conversion between ap and Kp indices is described at:
    https://www.ngdc.noaa.gov/stp/GEOMAG/kp_ap.html

    Assigns new data to '3hr_ap'.

    """

    # Kp are keys, where n.3 = n+ and n.6 = (n+1)-. E.g., 0.6 = 1-
    kp_to_ap = {0: 0, 0.3: 2, 0.6: 3, 1: 4, 1.3: 5, 1.6: 6, 2: 7, 2.3: 9,
                2.6: 12, 3: 15, 3.3: 18, 3.6: 22, 4: 27, 4.3: 32, 4.6: 39,
                5: 48, 5.3: 56, 5.6: 67, 6: 80, 6.3: 94, 6.6: 111, 7: 132,
                7.3: 154, 7.6: 179, 8: 207, 8.3: 236, 8.6: 300, 9: 400}

    def ap(kk):
        if np.isfinite(kk):
            return kp_to_ap[np.floor(kk * 10.0) / 10.0]
        else:
            return np.nan

    # Test the input
    if var_name not in kp_inst.variables:
        raise ValueError('Variable name for Kp data is missing: {:}'.format(
            var_name))

    # Convert from Kp to ap
    fill_val = kp_inst.meta[var_name][kp_inst.meta.labels.fill_val]
    ap_data = np.array([ap(kp) if kp != fill_val else fill_val
                        for kp in kp_inst[var_name]])

    # Append the output to the pysat instrument
    kp_inst['3hr_ap'] = pds.Series(ap_data, index=kp_inst.index)

    # Add metadata
    initialize_ap_metadata(kp_inst.meta, '3hr_ap', fill_val)
    kp_inst.meta['3hr_ap',
                 kp_inst.meta.labels.desc] = "3-hr ap (equivalent range) index"
    kp_inst.meta['3hr_ap', kp_inst.meta.labels.notes] = ''.join([
        'ap converted from Kp as described at: ',
        'https://www.ngdc.noaa.gov/stp/GEOMAG/kp_ap.html'])

    return


def calc_daily_Ap(ap_inst, ap_name='3hr_ap', daily_name='Ap',
                  running_name=None, min_periods=8):
    """Calculate the daily Ap index from the 3hr ap index.

    Parameters
    ----------
    ap_inst : pysat.Instrument
        pysat instrument containing 3-hourly ap data
    ap_name : str
        Column name for 3-hourly ap data (default='3hr_ap')
    daily_name : str
        Column name for daily Ap data (default='Ap')
    running_name : str or NoneType
        Column name for daily running average of ap, not output if None
        (default=None)
    min_periods : int
        Mininmum number of observations needed to output an average value
        (default=8)

    Raises
    ------
    ValueError
        If `ap_name` or `daily_name` aren't present in `ap_inst`

    Note
    ----
    Ap is the mean of the 3hr ap indices measured for a given day

    Option for running average is included since this information is used
    by MSIS when running with sub-daily geophysical inputs

    """

    # Test that the necessary data is available
    if ap_name not in ap_inst.variables:
        raise ValueError("bad 3-hourly ap column name: {:}".format(ap_name))

    # Test to see that we will not be overwritting data
    if daily_name in ap_inst.variables:
        raise ValueError("daily Ap column name already exists: " + daily_name)

    # Calculate the daily mean value
    ap_mean = ap_inst[ap_name].rolling(window='1D',
                                       min_periods=min_periods).mean()

    if running_name is not None:
        ap_inst[running_name] = ap_mean

        initialize_ap_metadata(ap_inst.meta, running_name,
                               ap_inst.meta[ap_name][
                                   ap_inst.meta.labels.fill_val])
        ap_inst.meta[running_name,
                     ap_inst.meta.labels.desc] = "running daily Ap index"
        ap_inst.meta[running_name,
                     ap_inst.meta.labels.notes] = ''.join(['24-h running ave',
                                                           ' of 3-hourly ',
                                                           'ap indices'])

    # Resample, backfilling so that each day uses the mean for the data from
    # that day only
    #
    # Pad the data so the first day will be backfilled
    ap_pad = pds.Series(np.full(shape=(1,), fill_value=np.nan),
                        index=[ap_mean.index[0] - pds.DateOffset(hours=3)])

    # Extract the mean that only uses data for one day
    ap_sel = ap_pad.combine_first(ap_mean[[i for i, tt in
                                           enumerate(ap_mean.index)
                                           if tt.hour == 21]])

    # Backfill this data
    ap_data = ap_sel.resample('3H').bfill()

    # Save the output for the original time range
    ap_inst[daily_name] = pds.Series(ap_data[1:], index=ap_data.index[1:])

    # Add metadata
    initialize_ap_metadata(ap_inst.meta, 'Ap',
                           ap_inst.meta[ap_name][ap_inst.meta.labels.fill_val])
    ap_inst.meta[daily_name, ap_inst.meta.labels.desc] = "daily Ap index"
    ap_inst.meta[daily_name,
                 ap_inst.meta.labels.notes] = ''.join(['Ap daily mean ',
                                                       'calculated from ',
                                                       '3-hourly ap indices'])
    return


def filter_geomag(inst, min_kp=0, max_kp=9, filter_time=24, kp_inst=None,
                  var_name='Kp'):
    """Filter pysat.Instrument data for given time after Kp drops below gate.

    Parameters
    ----------
    inst : pysat.Instrument or NoneType
        Instrument with non-Kp data to be filtered by geomagnetic activity
    min_kp : float
        Minimum Kp value allowed. Kp values below this filter the data in
        inst (default=0)
    max_kp : float
        Maximum Kp value allowed. Kp values above this filter the data in
        inst (default=9)
    filter_time : int
        Number of hours to filter data after Kp drops below max_kp (default=24)
    kp_inst : pysat.Instrument or NoneType
        Kp pysat.Instrument object with or without data already loaded. If None,
        will load GFZ definitive kp data for the instrument date (default=None)
    var_name : str
        String providing the variable name for the Kp data (default='Kp')

    Raises
    ------
    IOError
        If no Kp data is available to load

    Note
    ----
    Loads Kp data for the same timeframe covered by inst and sets inst.data to
    NaN for times when Kp > max_kp or Kp < min_kp and for filter_time after Kp
    drops below max_kp.

    Default max and min values accept all Kp, so changing only one will cause
    the filter to act as a high- or low-pass function.

    This routine is written for standard Kp data (tag=''), not the forecast or
    recent data.  However, it will work with these Kp data if they are supplied.

    """
    # Load the desired data
    if kp_inst is None:
        kp_inst = pysat.Instrument(inst_module=pysat_sw.instruments.sw_kp,
                                   tag='def', pad=dt.timedelta(days=1))

    if kp_inst.empty:
        load_kwargs = {'date': inst.index[0], 'end_date': inst.index[-1],
                       'verifyPad': True}

        # TODO(#131): Remove version check after minimum version supported
        # is 3.2.0
        if all([Version(pysat.__version__) > Version('3.0.1'),
                Version(pysat.__version__) < Version('3.2.0')]):
            load_kwargs['use_header'] = True

        kp_inst.load(**load_kwargs)

    if kp_inst.empty:
        raise IOError(
            'unable to load {:} data for {:}-{:}, check local data'.format(
                ':'.join([inst.platform, inst.name, inst.tag, inst.inst_id]),
                inst.index[0], inst.index[-1]))

    # Begin filtering, starting at the beginning of the instrument data
    sel_data = kp_inst[(inst.index[0] - dt.timedelta(days=1)):
                       (inst.index[-1] + dt.timedelta(days=1))]
    ind, = np.where((sel_data[var_name] > max_kp)
                    | (sel_data[var_name] < min_kp))

    for lind in ind:
        # Determine the time filter range for removing each flagged data
        sind = sel_data.index[lind]
        eind = sind + pds.DateOffset(hours=filter_time)
        inst[sind:eind] = np.nan
        inst.data = inst.data.dropna(axis=0, how='all')

    return


# --------------------------------------------------------------------------
# Common analysis functions

def convert_ap_to_kp(ap_data, fill_val=-1, ap_name='ap', kp_name='Kp'):
    """Convert Ap into Kp.

    Parameters
    ----------
    ap_data : array-like
        Array-like object containing Ap data
    fill_val : int, float, NoneType
        Fill value for the data set (default=-1)
    ap_name : str
        Name of the input ap (default='ap')
    kp_name : str
        Name of the output Kp (default='Kp')

    Returns
    -------
    kp_data : array-like
        Array-like object containing Kp data
    meta : Metadata
        Metadata object containing information about transformed data

    """

    # Convert from ap to Kp
    kp_data = np.array([round_ap(aa, fill_val=fill_val) for aa in ap_data])

    # Set the metadata
    meta = pysat.Meta()
    initialize_kp_metadata(meta, 'Kp', fill_val)
    meta['Kp', meta.labels.desc] = 'Kp converted from {:}'.format(ap_name)
    meta['Kp', meta.labels.notes] = ''.join(
        ['Kp converted from ', ap_name, 'as described at: ',
         'https://www.ngdc.noaa.gov/stp/GEOMAG/kp_ap.html'])

    # Return data and metadata
    return kp_data, meta


def round_ap(ap_in, fill_val=np.nan):
    """Round an ap value to the nearest Kp value.

    Parameters
    ----------
    ap_in : float
        Ap value as a floating point number.
    fill_val : float
        Value for unassigned or bad indices. (default=np.nan)

    Returns
    -------
    float
        Fill value for infinite or out-of-range data, otherwise the best kp
        index is provided.

    """

    # Define the ap to kp conversion
    one_third = 1.0 / 3.0
    two_third = 2.0 / 3.0
    ap_to_kp = {0: 0, 2: one_third, 3: two_third, 4: 1, 5: 1.0 + one_third,
                6: 1.0 + two_third, 7: 2, 9: 2.0 + one_third,
                12: 2.0 + two_third, 15: 3, 18: 3.0 + one_third,
                22: 3.0 + two_third, 27: 4, 32: 4.0 + one_third,
                39: 4.0 + two_third, 48: 5, 56: 5.0 + one_third,
                67: 5.0 + two_third, 80: 6, 94: 6.0 + one_third,
                111: 6.0 + two_third, 132: 7, 154: 7.0 + one_third,
                179: 7.0 + two_third, 207: 8, 236: 8.0 + one_third,
                300: 8.0 + two_third, 400: 9}
    max_ap = 400

    # Infinite or NaN values will return the fill value
    if not np.isfinite(ap_in):
        return fill_val

    # Ap values that correspond exactly to a Kp value will return that value
    if ap_in in ap_to_kp.keys():
        return ap_to_kp[ap_in]

    # The input Ap value may be appropriate, but does not correspond directly
    # to a Kp index.  Get a sorted list of directly corresponding Ap indices,
    # with Kp returned as double (N- = N.6667, N+ = N.3333333)
    ap_keys = sorted([akey for akey in ap_to_kp.keys() if akey <= ap_in])

    # If the value is too large or too small, return the fill value
    if len(ap_keys) == 0 or (ap_keys[-1] < ap_in and ap_keys[-1] == max_ap):
        return fill_val

    # The value is realistic, return the Kp value
    return ap_to_kp[ap_keys[-1]]


def combine_kp(standard_inst=None, recent_inst=None, forecast_inst=None,
               start=None, stop=None, fill_val=np.nan):
    """Combine the output from the different Kp sources for a range of dates.

    Parameters
    ----------
    standard_inst : pysat.Instrument or NoneType
        Instrument object containing data for the 'sw' platform, 'kp' name,
        and 'def' tag or None to exclude (default=None)
    recent_inst : pysat.Instrument or NoneType
        Instrument object containing data for the 'sw' platform, 'kp' name,
        and 'recent' tag or None to exclude (default=None)
    forecast_inst : pysat.Instrument or NoneType
        Instrument object containing data for the 'sw' platform, 'kp' name,
        and 'forecast' tag or None to exclude (default=None)
    start : dt.datetime or NoneType
        Starting time for combining data, or None to use earliest loaded
        date from the pysat Instruments (default=None)
    stop : dt.datetime
        Ending time for combining data, or None to use the latest loaded date
        from the pysat Instruments (default=None)
    fill_val : int or float
        Desired fill value (since the standard instrument fill value differs
        from the other sources) (default=np.nan)

    Returns
    -------
    kp_inst : pysat.Instrument
        Instrument object containing Kp observations for the desired period of
        time, merging the standard, recent, and forecasted values based on
        their reliability

    Raises
    ------
    ValueError
        If only one Kp instrument or bad times are provided

    Note
    ----
    Merging prioritizes the standard data, then the recent data, and finally
    the forecast data

    Will not attempt to download any missing data, but will load data

    """

    notes = "Combines data from"

    # Create an ordered list of the Instruments, excluding any that are None
    all_inst = list()
    tag = 'combined'
    inst_flag = None
    if standard_inst is not None:
        all_inst.append(standard_inst)
        tag += '_standard'
        if inst_flag is None:
            inst_flag = 'standard'

    if recent_inst is not None:
        all_inst.append(recent_inst)
        tag += '_recent'
        if inst_flag is None:
            inst_flag = 'recent'

    if forecast_inst is not None:
        all_inst.append(forecast_inst)
        tag += '_forecast'
        if inst_flag is None:
            inst_flag = 'forecast'

    if len(all_inst) < 2:
        raise ValueError("need at two Kp Instrument objects to combine them")

    # If the start or stop times are not defined, get them from the Instruments
    if start is None:
        stimes = [inst.index.min() for inst in all_inst if len(inst.index) > 0]
        start = min(stimes) if len(stimes) > 0 else None

    if stop is None:
        stimes = [inst.index.max() for inst in all_inst if len(inst.index) > 0]
        stop = max(stimes) + dt.timedelta(days=1) if len(stimes) > 0 else None

    if start is None or stop is None:
        raise ValueError(' '.join(("must either load in Instrument objects or",
                                   "provide starting and ending times")))

    # Initialize the output instrument
    # TODO(#136): Remove if/else when pysat is 3.2.0+
    if hasattr(all_inst[0], "meta_labels"):
        meta_kwargs = {"labels": all_inst[0].meta_labels}
        kp_inst = pysat.Instrument(labels=all_inst[0].meta_labels)
    else:
        meta_kwargs = all_inst[0].meta_kwargs
        kp_inst = pysat.Instrument(meta_kwargs=meta_kwargs)

    kp_inst.inst_module = pysat_sw.instruments.sw_kp
    kp_inst.tag = tag
    kp_inst.date = start
    kp_inst.doy = np.int64(start.strftime("%j"))
    kp_inst.meta = pysat.Meta(**meta_kwargs)
    initialize_kp_metadata(kp_inst.meta, 'Kp', fill_val=fill_val)

    kp_times = list()
    kp_values = list()

    # Cycle through the desired time range
    itime = start
    while itime < stop and inst_flag is not None:
        # Load and save the standard data for as many times as possible
        if inst_flag == 'standard':
            load_kwargs = {'date': itime}

            # TODO(#131): Remove version check after minimum version supported
            # is 3.2.0
            if all([Version(pysat.__version__) > Version('3.0.1'),
                    Version(pysat.__version__) < Version('3.2.0')]):
                load_kwargs['use_header'] = True

            standard_inst.load(**load_kwargs)

            if notes.find("standard") < 0:
                notes += " the {:} source ({:} to ".format(inst_flag,
                                                           itime.date())

            if len(standard_inst.index) == 0:
                inst_flag = 'forecast' if recent_inst is None else 'recent'
                notes += "{:})".format(itime.date())
            else:
                kp_times.extend(list(standard_inst.index))
                kp_values.extend(list(standard_inst['Kp']))
                itime = kp_times[-1] + pds.DateOffset(hours=3)

        # Load and save the recent data for as many times as possible
        if inst_flag == 'recent':
            # Determine which files should be loaded
            if len(recent_inst.index) == 0:
                files = np.unique(recent_inst.files.files[itime:stop])
            else:
                files = [None]  # No load needed, if already initialized

            # Cycle through all possible files of interest, saving relevant
            # data
            for filename in files:
                if filename is not None:
                    load_kwargs = {'fname': filename}

                    # TODO(#131): Remove version check after minimum version
                    # supported is 3.2.0
                    if all([Version(pysat.__version__) > Version('3.0.1'),
                            Version(pysat.__version__) < Version('3.2.0')]):
                        load_kwargs['use_header'] = True
                    recent_inst.load(**load_kwargs)

                if notes.find("recent") < 0:
                    notes += " the {:} source ({:} to ".format(inst_flag,
                                                               itime.date())

                # Determine which times to save
                local_fill_val = recent_inst.meta[
                    'Kp', recent_inst.meta.labels.fill_val]
                good_times = ((recent_inst.index >= itime)
                              & (recent_inst.index < stop))
                good_vals = recent_inst['Kp'][good_times] != local_fill_val

                # Save output data and cycle time
                kp_times.extend(list(recent_inst.index[good_times][good_vals]))
                kp_values.extend(list(recent_inst['Kp'][good_times][good_vals]))
                itime = kp_times[-1] + pds.DateOffset(hours=3)

            inst_flag = 'forecast' if forecast_inst is not None else None
            notes += "{:})".format(itime.date())

        # Load and save the forecast data for as many times as possible
        if inst_flag == "forecast":
            # Determine which files should be loaded
            if len(forecast_inst.index) == 0:
                files = np.unique(forecast_inst.files.files[itime:stop])
            else:
                files = [None]  # No load needed, if already initialized

            # Cycle through all possible files of interest, saving relevant
            # data
            for filename in files:
                if filename is not None:
                    load_kwargs = {'fname': filename}

                    # TODO(#131): Remove version check after minimum version
                    # supported is 3.2.0
                    if all([Version(pysat.__version__) > Version('3.0.1'),
                            Version(pysat.__version__) < Version('3.2.0')]):
                        load_kwargs['use_header'] = True
                    forecast_inst.load(**load_kwargs)

                if notes.find("forecast") < 0:
                    notes += " the {:} source ({:} to ".format(inst_flag,
                                                               itime.date())

                # Determine which times to save
                local_fill_val = forecast_inst.meta[
                    'Kp', forecast_inst.meta.labels.fill_val]
                good_times = ((forecast_inst.index >= itime)
                              & (forecast_inst.index < stop))
                good_vals = forecast_inst['Kp'][good_times] != local_fill_val

                # Save desired data and cycle time
                new_times = list(forecast_inst.index[good_times][good_vals])
                kp_times.extend(new_times)
                new_vals = list(forecast_inst['Kp'][good_times][good_vals])
                kp_values.extend(new_vals)
                itime = kp_times[-1] + pds.DateOffset(hours=3)
            notes += "{:})".format(itime.date())

            inst_flag = None

    if inst_flag is not None:
        notes += "{:})".format(itime.date())

    # Determine if the beginning or end of the time series needs to be padded
    freq = None if len(kp_times) < 2 else pysat.utils.time.calc_freq(kp_times)
    end_date = stop - pds.DateOffset(days=1)
    date_range = pds.date_range(start=start, end=end_date, freq=freq)

    if len(kp_times) == 0:
        kp_times = date_range

    if date_range[0] < kp_times[0]:
        # Extend the time and value arrays from their beginning with fill
        # values
        itime = abs(date_range - kp_times[0]).argmin()
        kp_times.reverse()
        kp_values.reverse()
        extend_times = list(date_range[:itime])
        extend_times.reverse()
        kp_times.extend(extend_times)
        kp_values.extend([fill_val for kk in extend_times])
        kp_times.reverse()
        kp_values.reverse()

    if date_range[-1] > kp_times[-1]:
        # Extend the time and value arrays from their end with fill values
        itime = abs(date_range - kp_times[-1]).argmin() + 1
        extend_times = list(date_range[itime:])
        kp_times.extend(extend_times)
        kp_values.extend([fill_val for kk in extend_times])

    # Save output data
    kp_inst.data = pds.DataFrame(kp_values, columns=['Kp'], index=kp_times)

    # Resample the output data, filling missing values
    if (date_range.shape != kp_inst.index.shape
            or abs(date_range - kp_inst.index).max().total_seconds() > 0.0):
        kp_inst.data = kp_inst.data.resample(freq).fillna(method=None)
        if np.isfinite(fill_val):
            kp_inst.data[np.isnan(kp_inst.data)] = fill_val

    # Update the metadata notes for this procedure
    notes += ", in that order"
    kp_inst.meta['Kp', kp_inst.meta.labels.notes] = notes

    return kp_inst



#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------
"""Supports Kp index values.

Properties
----------
platform
    'sw'
name
    'kp'
tag
    - '' Deprecated, mixed definitive and nowcast Kp data from GFZ
    - 'def' Definitive Kp data from GFZ
    - 'now' Nowcast Kp data from GFZ
    - 'prediction' Predictions from SWPC for the next 3 days
    - 'forecast' Forecast data from SWPC (next 3 days)
    - 'recent' The last 30 days of Kp data from SWPC
inst_id
    - ''

Note
----
Downloads data from ftp.gfz-potsdam.de or SWPC. These files also contain ap
data (with the GFZ data additionally containing Cp data), and so the additional
data files will be saved to the appropriate data directories to avoid multiple
downloads.

The historic definitive and nowcast Kp files are stored in yearly files, with
the current year being updated remotely on a regular basis.  If you are using
historic data for the current year, we recommend re-downloading it before
performing your data processing.

The forecast data is stored by generation date, where each file contains the
forecast for the next three days. Forecast data downloads are only supported
for the current day. When loading forecast data, the date specified with the
load command is the date the forecast was generated. The data loaded will span
three days. To always ensure you are loading the most recent data, load
the data with tomorrow's date.

Recent data is also stored by the generation date from the SWPC. Each file
contains 30 days of Kp measurements. The load date issued to pysat corresponds
to the generation date.

Examples
--------
::

    kp = pysat.Instrument('sw', 'kp', tag='recent')
    kp.download()
    kp.load(date=kp.tomorrow())


Warnings
--------
The 'forecast', 'prediction', and 'recent' tags load Kp data for a specific
period of time. Loading multiple files, loading multiple days, the data padding
feature, and multi_file_day feature available from the pyast.Instrument object
is not appropriate for these tags data.

"""

import datetime as dt
import numpy as np
import pandas as pds
import warnings

import pysat

from pysatSpaceWeather.instruments import methods

# ----------------------------------------------------------------------------
# Instrument attributes

platform = 'sw'
name = 'kp'
tags = {'': 'Deprecated, mixed definitive and nowcast Kp data from GFZ',
        'def': 'Definitive Kp data from GFZ',
        'now': 'Nowcast Kp data from GFZ',
        'prediction': 'SWPC Predictions for the next three days',
        'forecast': 'SWPC Forecast data next (3 days)',
        'recent': 'SWPC provided Kp for past 30 days'}
inst_ids = {'': list(tags.keys())}

# Generate todays date to support loading forecast data
now = dt.datetime.utcnow()
today = dt.datetime(now.year, now.month, now.day)

# ----------------------------------------------------------------------------
# Instrument test attributes

# Set test dates
_test_dates = {'': {'def': dt.datetime(2009, 1, 1),
                    'now': dt.datetime(2020, 1, 1),
                    'forecast': today + dt.timedelta(days=1),
                    'prediction': today + dt.timedelta(days=1),
                    'recent': today}}

# Other tags assumed to be True
_test_download_ci = {'': {'': False}}

# ----------------------------------------------------------------------------
# Instrument methods

preprocess = methods.general.preprocess


def init(self):
    """Initialize the Instrument object with instrument specific values."""

    self.acknowledgements = methods.kp_ap.acknowledgements(self.name, self.tag)
    self.references = methods.kp_ap.references(self.name, self.tag)
    pysat.logger.info(self.acknowledgements)

    return


def clean(self):
    """Clean the Kp, not required for this index (empty function)."""

    return


# ----------------------------------------------------------------------------
# Instrument functions


def load(fnames, tag='', inst_id=''):
    """Load Kp index files.

    Parameters
    ----------
    fnames : pandas.Series
        Series of filenames
    tag : str
        Instrument tag (default='')
    inst_id : str
        Instrument ID, not used. (default='')

    Returns
    -------
    result : pandas.DataFrame
        Object containing satellite data
    meta : pysat.Meta
        Object containing metadata such as column names and units

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    Warnings
    --------
    tag '' has been deprecated, will be removed in version 0.2.0+

    """

    meta = pysat.Meta()
    if tag == '':
        # This data type has been deprecated due to changes at GFZ
        warnings.warn("".join(["Changes at the GFZ database have led to this",
                               " data type being deprecated. Switch to using",
                               " 'def' for definitive Kp or 'now' for Kp ",
                               "nowcasts from GFZ. Load support will be ",
                               "removed in version 0.2.0+"]),
                      DeprecationWarning, stacklevel=2)

        # Kp data stored monthly, need to return data daily.  The daily date is
        # attached to filename.  Parse off the last date, load month of data,
        # and downselect to the desired day
        data = pds.DataFrame()

        # Set up fixed width format for these files, only selecting the date
        # and daily 3-hour Kp values
        date_slice = slice(0, 6)
        kp_slice = [slice(7, 10), slice(10, 13), slice(13, 16), slice(16, 19),
                    slice(19, 23), slice(23, 26), slice(26, 29), slice(29, 32)]

        # These are monthly files, if a date range is desired, test here.
        # Does not assume an ordered list, but the date range must be continous
        # within a given month.
        unique_fnames = dict()
        for filename in fnames:
            fname = filename[0:-11]
            fdate = dt.datetime.strptime(filename[-10:], '%Y-%m-%d')
            if fname not in unique_fnames.keys():
                unique_fnames[fname] = [fdate]
            else:
                unique_fnames[fname].append(fdate)

        # Load all of the desired filenames
        all_data = []
        for fname in unique_fnames.keys():
            # The daily date is attached to the filename.  Parse off the last
            # date, load month of data, downselect to the desired day
            fdate = min(unique_fnames[fname])

            with open(fname, 'r') as fin:
                temp = fin.readlines()

            if len(temp) == 0:
                pysat.logger.warn('Empty file: {:}'.format(fname))
                continue

            # This file has a different format if it is historic or a file that
            # is being actively updated.  In either case, this line will
            # remove the appropriate number of summmary lines.
            ilast = -1 if temp[-1].find('Mean') > 0 else -4
            temp = np.array(temp[:ilast])

            # Re-format the time data
            temp_index = np.array([dt.datetime.strptime(tline[date_slice],
                                                        '%y%m%d')
                                   for tline in temp])

            idx, = np.where((temp_index >= fdate)
                            & (temp_index < max(unique_fnames[fname])
                               + dt.timedelta(days=1)))

            temp_data = list()
            for tline in temp[idx]:
                temp_data.append(list())
                for col in kp_slice:
                    temp_data[-1].append(tline[col].strip())

            # Select the desired times and add to data list
            all_data.append(pds.DataFrame(temp_data, index=temp_index[idx]))

        # Combine data together
        data = pds.concat(all_data, axis=0, sort=True)

        if len(data.index) > 0:
            # Each column increments UT by three hours. Produce a single data
            # series that has Kp value monotonically increasing in time with
            # appropriate datetime indices
            data_series = pds.Series(dtype='float64')
            for i in np.arange(8):
                tind = data.index + pds.DateOffset(hours=int(3 * i))
                temp = pds.Series(data.iloc[:, i].values, index=tind)
                data_series = pds.concat([data_series, temp])

            data_series = data_series.sort_index()
            data_series.index.name = 'time'

            # Kp comes in non-user friendly values like 2-, 2o, and 2+. Relate
            # these to 1.667, 2.0, 2.333 for processing and user friendliness
            first = np.array([np.float64(str_val[0])
                              for str_val in data_series])
            flag = np.array([str_val[1] for str_val in data_series])

            ind, = np.where(flag == '+')
            first[ind] += 1.0 / 3.0
            ind, = np.where(flag == '-')
            first[ind] -= 1.0 / 3.0

            result = pds.DataFrame(first, columns=['Kp'],
                                   index=data_series.index)
        else:
            result = pds.DataFrame()

        fill_val = np.nan
    elif tag in ['def', 'now']:
        # Load the definitive or nowcast data. The Kp data stored in yearly
        # files, and we need to return data daily.  The daily date is
        # attached to filename.  Parse off the last date, load month of data,
        # and downselect to the desired day
        result = methods.gfz.load_def_now(fnames)
        fill_val = np.nan
    else:
        # Load the prediction, forecast or recent data
        all_data = []
        for fname in fnames:
            result = pds.read_csv(fname, index_col=0, parse_dates=True)
            all_data.append(result)

        result = pds.concat(all_data)
        fill_val = -1

    # Initalize the meta data
    if tag in ['', 'forecast', 'recent', 'prediction']:
        for kk in result.keys():
            methods.kp_ap.initialize_kp_metadata(meta, kk, fill_val)
    else:
        for kk in result.keys():
            if kk.find('Kp') >= 0:
                methods.kp_ap.initialize_kp_metadata(meta, kk, fill_val)
            elif kk.find('Bartels') >= 0:
                methods.kp_ap.initialize_bartel_metadata(meta, kk)

    return result, meta


def list_files(tag='', inst_id='', data_path='', format_str=None):
    """List local files for the requested Instrument tag.

    Parameters
    -----------
    tag : str
        Instrument tag, accepts any value from `tags`. (default='')
    inst_id : str
        Instrument ID, not used. (default='')
    data_path : str
        Path to data directory. (default='')
    format_str : str or NoneType
        User specified file format.  If None is specified, the default
        formats associated with the supplied tags are used. (default=None)

    Returns
    -------
    files : pysat._files.Files
        A class containing the verified available files

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    Warnings
    --------
    The '' tag has been deprecated and local file listing support will
    be removed in version 0.2.0+

    """

    if tag == '':
        # This data type has been deprecated due to changes at GFZ
        warnings.warn("".join(["Changes at the GFZ database have led to this",
                               " data type being deprecated. Switch to using",
                               " 'def' for definitive Kp or 'now' for Kp ",
                               "nowcasts from GFZ. Local file listing support ",
                               "will be removed in version 0.2.0+"]),
                      DeprecationWarning, stacklevel=2)

        # Files are by month, going to add date to monthly filename for
        # each day of the month. The load routine will load a month of
        # data and use the appended date to select out appropriate data.
        if format_str is None:
            format_str = 'kp{year:2d}{month:02d}.tab'
        files = pysat.Files.from_os(data_path=data_path, format_str=format_str,
                                    two_digit_year_break=0)
        if not files.empty:
            files.loc[files.index[-1] + pds.DateOffset(months=1)
                      - pds.DateOffset(days=1)] = files.iloc[-1]
            files = files.asfreq('D', 'pad')
            files = files + '_' + files.index.strftime('%Y-%m-%d')
    elif tag in ['def', 'now']:
        files = methods.gfz.kp_ap_cp_list_files(name, tag, inst_id, data_path,
                                                format_str=format_str)
    else:
        files = methods.swpc.list_files(name, tag, inst_id, data_path,
                                        format_str=format_str)

    return files


def download(date_array, tag, inst_id, data_path, mock_download_dir=None):
    """Download the Kp index data from the appropriate repository.

    Parameters
    ----------
    date_array : array-like or pandas.DatetimeIndex
        Array-like or index of datetimes to be downloaded.
    tag : str
        Denotes type of file to load.
    inst_id : str
        Specifies the instrument identification, not used.
    data_path : str
        Path to data directory.
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date as if they were
        downloaded (default=None)

    Raises
    ------
    IOError
        If an unknown mock download directory is supplied or there is a problem
        downloading the data.

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    Warnings
    --------
    Only able to download current forecast data, not archived forecasts.

    The '' tag has been deprecated and downloads are no longer supported by
    the source.  Use 'dep' or 'now' instead.

    """

    # Download standard Kp data
    if tag == '':
        # This data type has been deprecated due to changes at GFZ
        warnings.warn("".join(["Changes at the GFZ database have led to this",
                               " data type being deprecated. Switch to using",
                               " 'def' for definitive Kp or 'now' for Kp ",
                               "nowcasts from GFZ. Downloads are no longer ",
                               "supported by GFZ."]),
                      DeprecationWarning, stacklevel=2)
    elif tag in ['def', 'now']:
        methods.gfz.kp_ap_cp_download(platform, name, date_array, tag, inst_id,
                                      data_path, mock_download_dir)
    elif tag == 'forecast':
        methods.swpc.geomag_forecast_download(name, date_array, data_path,
                                              mock_download_dir)
    elif tag == 'recent':
        methods.swpc.kp_ap_recent_download(name, date_array, data_path,
                                           mock_download_dir)
    elif tag == 'prediction':
        methods.swpc.solar_geomag_predictions_download(
            name, date_array, data_path, mock_download_dir)

    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------
"""Supports the auroral electrojet AE values.

Properties
----------
platform
    'sw'
name
    'ae'
tag
    - 'lasp' Predicted AE from real-time ACE or DSCOVR provided by LASP
inst_id
    - ''

"""

import datetime as dt
import numpy as np

import pysat

from pysatSpaceWeather.instruments.methods import auroral_electrojet as mm_ae
from pysatSpaceWeather.instruments.methods import lasp

# ----------------------------------------------------------------------------
# Instrument attributes

platform = 'sw'
name = 'ae'
tags = {'lasp': 'Predicted AE from real-time ACE or DSCOVR provided by LASP'}
inst_ids = {'': [tag for tag in tags.keys()]}

# Generate today's date to support loading predicted data sets
today = pysat.utils.time.today()
tomorrow = today + dt.timedelta(days=1)

# ----------------------------------------------------------------------------
# Instrument test attributes

_test_dates = {'': {'lasp': today}}

# ----------------------------------------------------------------------------
# Instrument methods


def init(self):
    """Initialize the Instrument object with instrument specific values."""

    self.acknowledgements = mm_ae.acknowledgements(self.name, self.tag)
    self.references = mm_ae.references(self.name, self.tag)
    pysat.logger.info(self.acknowledgements)
    return


def clean(self):
    """Clean the AE index, empty function."""
    return


# ----------------------------------------------------------------------------
# Instrument functions


def load(fnames, tag='', inst_id=''):
    """Load the AE index files.

    Parameters
    ----------
    fnames : pandas.Series
        Series of filenames
    tag : str
        Instrument tag string. (default='')
    inst_id : str
        Instrument ID, not used. (default='')

    Returns
    -------
    data : pandas.DataFrame
        Object containing satellite data
    pysat.Meta
        Object containing metadata such as column names and units

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """
    data = pysat.instruments.methods.general.load_csv_data(
        fnames, read_csv_kwargs={'index_col': 0, 'parse_dates': True})

    # Create metadata
    meta = pysat.Meta()
    meta['ae'] = {meta.labels.units: 'nT',
                  meta.labels.name: 'AE',
                  meta.labels.notes: tags[tag],
                  meta.labels.desc: ''.join([
                      'Auroral Electrojet index, best estimate ',
                      'of the average value over the next two hours']),
                  meta.labels.fill_val: np.nan,
                  meta.labels.min_val: -np.inf,
                  meta.labels.max_val: np.inf}

    return data, meta


def list_files(tag='', inst_id='', data_path='', format_str=None):
    """List local data files for AE data.

    Parameters
    ----------
    tag : str
        Instrument tag, accepts any value from `tags`. (default='')
    inst_id : str
        Instrument ID, not used. (default='')
    data_path : str
        Path to data directory. (default='')
    format_str : str or NoneType
        User specified file format.  If None is specified, the default
        formats associated with the supplied tags are used. (default=None)

    Returns
    -------
    files : pysat.Files
        A class containing the verified available files

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """
    # Get the format string, if not supplied by the user
    if format_str is None:
        format_str = ''.join(['sw_ae_', tag, '_{year:4d}-{month:2d}-',
                              '{day:2d}.txt'])

    # Get the desired files
    files = pysat.Files.from_os(data_path=data_path, format_str=format_str)

    return files


def download(date_array, tag, inst_id, data_path, mock_download_dir=None):
    """Download the AE index data from the appropriate repository.

    Parameters
    ----------
    date_array : array-like or pandas.DatetimeIndex
        Array-like or index of datetimes for which files will be downloaded.
    tag : str
        Instrument tag, used to determine download location.
    inst_id : str
        Instrument ID, not used.
    data_path : str
        Path to data directory.
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date as if they were
        downloaded (default=None)

    Raises
    ------
    IOError
        If the data link has an unexpected format or an unknown mock download
        directory is supplied.

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """
    lasp.prediction_downloads(name, tag, data_path, mock_download_dir)

    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------
"""Supports ACE Solar Isotope Spectrometer data.

Properties
----------
platform
    'ace' Advanced Composition Explorer
name
    'sis' Solar Isotope Spectrometer
tag
    - 'realtime' Real-time data from the Space Weather Prediction Center (SWPC)
    - 'historic' Historic data from the SWPC
inst_id
    - ''

Note
----
This is not the ACE scientific data set, which will be available at pysatNASA

Examples
--------
The real-time data is stored by generation date, where each file contains the
data for the current day.  If you leave download dates empty, though, it will
grab today's file three times and assign dates from yesterday, today, and
tomorrow.
::

    sis = pysat.Instrument('ace', 'sis', tag='realtime')
    sis.download(start=sis.today())
    sis.load(date=sis.today())



Warnings
--------
The 'realtime' data contains a changing period of time. Loading multiple files,
loading multiple days, the data padding feature, and multi_file_day feature
available from the pyast.Instrument object is not appropriate for 'realtime'
data.

"""

import datetime as dt
import functools
import numpy as np

from pysat.instruments.methods.general import load_csv_data
from pysat import logger

from pysatSpaceWeather.instruments.methods import ace as mm_ace
from pysatSpaceWeather.instruments.methods import general

# ----------------------------------------------------------------------------
# Instrument attributes

platform = 'ace'
name = 'sis'
tags = {'realtime': 'Real-time data from the SWPC',
        'historic': ' Historic data from the SWPC'}
inst_ids = {inst_id: [tag for tag in tags.keys()] for inst_id in ['']}

# Define today's date
now = dt.datetime.utcnow()

# ----------------------------------------------------------------------------
# Instrument test attributes

# Set test dates (first level: inst_id, second level: tag)
_test_dates = {inst_id: {'realtime': dt.datetime(now.year, now.month, now.day),
                         'historic': dt.datetime(2009, 1, 1)}
               for inst_id in inst_ids.keys()}

# Set clean warning tests
_clean_warn = {inst_id: {tag: mm_ace.clean_warn for tag in inst_ids[inst_id]}
               for inst_id in inst_ids.keys()}

# ----------------------------------------------------------------------------
# Instrument methods

preprocess = general.preprocess


def init(self):
    """Initialize the Instrument object with instrument specific values."""

    # Set the appropraite acknowledgements and references
    self.acknowledgements = mm_ace.acknowledgements()
    self.references = mm_ace.references(self.name)

    logger.info(self.acknowledgements)

    return


def clean(self):
    """Clean real-time ACE data using the status flag.

    Note
    ----
    Supports 'clean' and 'dirty'.  Replaces all fill values with NaN.
    Clean - status flag of zero (nominal data)
    Dirty - status flag < 9 (accepts bad data record, removes no data record)

    """
    # Perform the standard ACE cleaning
    max_status = mm_ace.clean(self)

    # Evaluate the different proton fluxes. Replace bad values with NaN and
    # times with no valid data
    self.data['int_pflux_10MeV'][self.data['status_10'] > max_status] = np.nan
    self.data['int_pflux_30MeV'][self.data['status_30'] > max_status] = np.nan

    eval_cols = ['int_pflux_10MeV', 'int_pflux_30MeV']

    # Remove lines without any good data
    good_cols = (np.isfinite(self.data.loc[:, eval_cols])).sum(axis=1)
    bad_index = good_cols[good_cols == 0].index
    self.data = self.data.drop(index=bad_index)

    return


# ----------------------------------------------------------------------------
# Instrument functions

download = functools.partial(mm_ace.download, name=name, now=now)
list_files = functools.partial(mm_ace.list_files, name=name)


def load(fnames, tag='', inst_id=''):
    """Load the ACE space weather prediction data.

    Parameters
    ----------
    fnames : array-like
        Series, list, or array of filenames.
    tag : str
        Instrument tag, not used. (default='')
    inst_id : str
        ACE instrument ID, not used. (default='')

    Returns
    -------
    data : pandas.DataFrame
        Object containing instrument data
    meta : pysat.Meta
        Object containing metadata such as column names and units

    See Also
    --------
    pysat.instruments.methods.general.load_csv_data

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """

    # Save each file to the output DataFrame
    data = load_csv_data(fnames, read_csv_kwargs={'index_col': 0,
                                                  'parse_dates': True})

    # Assign the meta data
    meta, status_desc = mm_ace.common_metadata()
    flux_name = 'Integral Proton Flux'

    meta['status_10'] = {meta.labels.units: '',
                         meta.labels.name: ''.join([flux_name,
                                                    ' > 10 MeV Status']),
                         meta.labels.notes: '',
                         meta.labels.desc: status_desc,
                         meta.labels.fill_val: np.nan,
                         meta.labels.min_val: 0,
                         meta.labels.max_val: 9}
    meta['status_30'] = {meta.labels.units: '',
                         meta.labels.name: ''.join([flux_name,
                                                    ' > 30 MeV Status']),
                         meta.labels.notes: '',
                         meta.labels.desc: status_desc,
                         meta.labels.fill_val: np.nan,
                         meta.labels.min_val: 0,
                         meta.labels.max_val: 9}
    meta['int_pflux_10MeV'] = {meta.labels.units: 'p/cs2-sec-ster',
                               meta.labels.name: ''.join([flux_name,
                                                          ' > 10 MeV']),
                               meta.labels.notes: '',
                               meta.labels.desc: ''.join(['5-min averaged ',
                                                          flux_name,
                                                          ' > 10 MeV']),
                               meta.labels.fill_val: -1.0e5,
                               meta.labels.min_val: -np.inf,
                               meta.labels.max_val: np.inf}
    meta['int_pflux_30MeV'] = {meta.labels.units: 'p/cs2-sec-ster',
                               meta.labels.name: ''.join([flux_name,
                                                          ' > 30 MeV']),
                               meta.labels.notes: '',
                               meta.labels.desc: ''.join(['5-min averaged ',
                                                          flux_name,
                                                          ' > 30 MeV']),
                               meta.labels.fill_val: -1.0e5,
                               meta.labels.min_val: -np.inf,
                               meta.labels.max_val: np.inf}

    return data, meta



#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------
"""Provides support routines for auroral electrojet indices."""


def acknowledgements(name, tag):
    """Define the acknowledgements for the index and data source.

    Parameters
    ----------
    name : str
        Name of the auroral electrojet index
    tag : str
        Tag of the auroral electrojet index

    Returns
    -------
    ackn : str
        Acknowledgements string associated with the appropriate auroral
        electrojet tag.

    """

    ackn = {'lasp': ''.join(['Preliminary ', name.upper(), ' predictions are ',
                             'provided by LASP, contact Xinlin Li for more ',
                             'details <xinlin.li@lasp.colorado.edu>'])}

    return ackn[tag]


def references(name, tag):
    """Define the references for the auroral electrojet data.

    Parameters
    ----------
    name : str
        Name of the auroral electrojet index
    tag : string
        Tag of the space weather index

    Returns
    -------
    refs : str
        Reference string associated with the appropriate Dst tag.

    """

    davis = ''.join(['Davis, T. N., and Sugiura, M. (1966), Auroral electrojet',
                     ' activity index AE and its universal time variations, ',
                     'J. Geophys. Res., 71( 3), 785– 801, ',
                     'doi:10.1029/JZ071i003p00785.'])

    luo_au_al_ae = ''.join(['Luo, B., Li, X., Temerin, M., and Liu, S. (2013),',
                            'Prediction of the AU, AL, and AE indices using ',
                            'solar wind parameters, J. Geophys. Res. Space ',
                            'Physics, 118, 7683– 7694, ',
                            'doi:10.1002/2013JA019188.'])
    li_al = ''.join(['Li, X., Oh, K. S., and Temerin, M. (2007), Prediction ',
                     'of the AL index using solar wind parameters, J. ',
                     'Geophys. Res., 112, A06224, doi:10.1029/2006JA011918.'])

    refs = {'lasp': {'ae': '\n'.join([davis, luo_au_al_ae]),
                     'au': '\n'.join([davis, luo_au_al_ae]),
                     'al': '\n'.join([davis, li_al, luo_au_al_ae])}}

    return refs[tag][name]



from pysatSpaceWeather.instruments.methods import ace  # noqa F401
from pysatSpaceWeather.instruments.methods import auroral_electrojet  # noqa F401
from pysatSpaceWeather.instruments.methods import dst  # noqa F401
from pysatSpaceWeather.instruments.methods import f107  # noqa F401
from pysatSpaceWeather.instruments.methods import general  # noqa F401
from pysatSpaceWeather.instruments.methods import gfz  # noqa F401
from pysatSpaceWeather.instruments.methods import kp_ap  # noqa F401
from pysatSpaceWeather.instruments.methods import lasp  # noqa F401
from pysatSpaceWeather.instruments.methods import lisird # noqa F401
from pysatSpaceWeather.instruments.methods import swpc  # noqa F401



#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------
"""Supports Hpo index values.

Properties
----------
platform
    'sw'
name
    'hpo'
tag
    - 'now' Near real-time nowcast Hpo data from GFZ
inst_id
    - '30min' Half-hourly resolution indices
    - '60min' Hourly resolution indices

Note
----
Downloads data from ftp.gfz-potsdam.de

Examples
--------
::

    hp30 = pysat.Instrument('sw', 'hpo', tag='now', inst_id='30min')
    hp30.download(2009, 1)
    hp30.load(yr=2009, doy=1)

"""

import datetime as dt
import numpy as np
import pandas as pds

import pysat

from pysatSpaceWeather.instruments import methods

# ----------------------------------------------------------------------------
# Instrument attributes

platform = 'sw'
name = 'hpo'
tags = {'now': 'Near real-time nowcast Hpo data from GFZ'}
inst_ids = {inst_id: list(tags.keys()) for inst_id in ['30min', '60min']}

# ----------------------------------------------------------------------------
# Instrument test attributes

# Set test dates
_test_dates = {inst_id: {tag: dt.datetime(2009, 1, 1)
                         for tag in inst_ids[inst_id]}
               for inst_id in inst_ids.keys()}

# ----------------------------------------------------------------------------
# Instrument methods

preprocess = methods.general.preprocess


def init(self):
    """Initialize the Instrument object with instrument specific values."""

    self.acknowledgements = methods.gfz.ackn
    self.references = methods.gfz.hpo_refs
    pysat.logger.info(self.acknowledgements)
    return


def clean(self):
    """Clean the Kp, not required for this index (empty function)."""

    return


# ----------------------------------------------------------------------------
# Instrument functions


def load(fnames, tag='', inst_id=''):
    """Load Hpo index files.

    Parameters
    ----------
    fnames : pandas.Series
        Series of filenames.
    tag : str
        Instrument tag. (default='')
    inst_id : str
        Instrument ID, not used. (default='')

    Returns
    -------
    result : pandas.DataFrame
        Object containing satellite data
    meta : pysat.Meta
        Object containing metadata such as column names and units

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """

    # Load the nowcast data. The Hpo data is stored in monthly files, and we
    # need to return data daily.  The daily date is attached to filename.
    # Parse off the last date, load month of data, and downselect to the
    # desired day
    result = methods.gfz.load_def_now(fnames)

    # Initalize the meta data
    dkey = 'Hp{:s}'.format(inst_id.split('min')[0])
    meta = pysat.Meta()
    meta[dkey] = {meta.labels.units: '',
                  meta.labels.name: dkey,
                  meta.labels.desc: "{:s}ourly Planetary Open index".format(
                      'H' if inst_id == '60min' else 'Half-h'),
                  meta.labels.min_val: 0,
                  meta.labels.max_val: np.inf,
                  meta.labels.fill_val: np.nan}

    return result, meta


def list_files(tag='', inst_id='', data_path='', format_str=None):
    """List local files for the requested Instrument tag.

    Parameters
    -----------
    tag : str
        Instrument tag, accepts any value from `tags`. (default='')
    inst_id : str
        Instrument ID, not used. (default='')
    data_path : str
        Path to data directory. (default='')
    format_str : str or NoneType
        User specified file format.  If None is specified, the default
        formats associated with the supplied tags are used. (default=None)

    Returns
    -------
    files : pysat._files.Files
        A class containing the verified available files

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """

    if format_str is None:
        format_str = 'Hp{:s}_{{year:04d}}-{{month:02d}}.txt'.format(
            inst_id.split('min')[0])

    # Files are stored by month, going to add a date to the monthly filename for
    # each day of month.  The load routine will load the year and use the append
    # date to select out approriate data.
    files = pysat.Files.from_os(data_path=data_path, format_str=format_str)
    if not files.empty:
        files.loc[files.index[-1]
                  + pds.DateOffset(months=1, days=-1)] = files.iloc[-1]
        files = files.asfreq('D', 'pad')
        files = files + '_' + files.index.strftime('%Y-%m-%d')

    return files


def download(date_array, tag, inst_id, data_path, update_files=False,
             mock_download_dir=None):
    """Download the Hpo index data from the appropriate repository.

    Parameters
    ----------
    date_array : array-like or pandas.DatetimeIndex
        Array-like or index of datetimes to be downloaded.
    tag : str
        Denotes type of file to load.
    inst_id : str
        Specifies the instrument identification, not used.
    data_path : str
        Path to data directory.
    update_files : bool
        Re-download data for files that already exist if True (default=False)
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date as if they were
        downloaded (default=None)

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    Warnings
    --------
    Only able to download current recent data, not archived forecasts.

    Raises
    ------
    IOError
        If an unknown mock download directory is supplied.

    """

    # Set the download input options
    gfz_data_name = 'Hp{:s}'.format(inst_id.split('min')[0])
    local_file_prefix = '{:s}_'.format(gfz_data_name)
    local_date_fmt = "%Y-%m"
    freq = pds.DateOffset(months=1, seconds=-1)

    # Call the download routine
    methods.gfz.json_downloads(date_array, data_path, local_file_prefix,
                               local_date_fmt, gfz_data_name, freq,
                               update_files=update_files,
                               mock_download_dir=mock_download_dir)

    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------
"""Provides default routines for Dst."""


def acknowledgements(tag):
    """Define the acknowledgements for the Dst data.

    Parameters
    ----------
    tag : str
        Tag of the space weather index

    Returns
    -------
    ackn : str
        Acknowledgements string associated with the appropriate Dst tag.

    """

    ackn = {'noaa': 'Dst is maintained at NCEI (formerly NGDC) at NOAA',
            'lasp': ''.join(['Preliminary Dst predictions are provided by ',
                             'LASP, contact Xinlin Li for more details ',
                             '<xinlin.li@lasp.colorado.edu>'])}

    return ackn[tag]


def references(tag):
    """Define the references for the Dst data.

    Parameters
    ----------
    tag : string
        Tag of the space weather index

    Returns
    -------
    refs : str
        Reference string associated with the appropriate Dst tag.

    """

    refs = {'noaa': ''.join(['See referenece list and publication at: ',
                             'Sugiura M. and T. Kamei, http://',
                             'wdc.kugi.kyoto-u.ac.jp/dstdir/dst2/',
                             'onDstindex.html, last updated June 1991, ',
                             'accessed Dec 2020']),
            'lasp': ''.join(['A New Model for the Prediction of Dst on the ',
                             'Basis of the Solar Wind [Temerin and Li, 2002] ',
                             'and Dst model for 1995-2002 [Temerin and Li, ',
                             '2006]'])}

    return refs[tag]



#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------
"""Provides routines that support GFZ space weather instruments."""

import datetime as dt
import json
import numpy as np
import os
import pandas as pds
import requests

import pysat

from pysatSpaceWeather.instruments.methods import general

# ----------------------------------------------------------------------------
# Define the module variables

ackn = ''.join(['CC BY 4.0, This index is produced by the Geomagnetic ',
                'Observatory Niemegk, GFZ German Research Centre for ',
                'Geosciences.  Please cite the references in the ',
                "'references' attribute"])
geoind_refs = '\n'.join([''.join(["Bartels, J. (1949) - The standardized index",
                                  "Ks and the planetary index Kp, IATME ",
                                  "Bulletin 12b, 97."]),
                         ''.join(["Matzka, J., Bronkalla, O., Tornow, K., ",
                                  "Elger, K. and Stolle, C., 2021. ",
                                  "Geomagnetic Kp index. V. 1.0. GFZ Data ",
                                  "Services, doi:10.5880/Kp.0001"]),
                         ''.join(["Matzka, J., Stolle, C., Yamazaki, Y., ",
                                  "Bronkalla, O. and Morschhauser, A., 2021. ",
                                  "The geomagnetic Kp index and derived ",
                                  "indices of geomagnetic activity. Space ",
                                  "Weather,doi:10.1029/2020SW002641"])])
hpo_refs = '\n'.join([''.join(["Yamazaki, Y., Matzka, J., Stolle, C., ",
                               "Kervalishvili, G., Rauberg, J., Bronkalla, O.,",
                               " Morschhauser, A., Bruinsma, S., Shprits, ",
                               "Y.Y., Jackson, D.R., 2022. Geomagnetic ",
                               "Activity Index Hpo. Geophys. Res. Lett., 49, ",
                               "e2022GL098860, doi:10.1029/2022GL098860"]),
                      ''.join(["Matzka, J., Bronkalla, O., Kervalishvili, G.,",
                               " Rauberg, J. and Yamazaki, Y., 2022. ",
                               "Geomagnetic Hpo index. V. 2.0. GFZ Data ",
                               "Services, doi:10.5880/Hpo.0002"])])


# ----------------------------------------------------------------------------
# Define the module functions

def json_downloads(date_array, data_path, local_file_prefix, local_date_fmt,
                   gfz_data_name, freq, update_files=False, is_def=False,
                   mock_download_dir=None):
    """Download data from GFZ into CSV files at a specified cadence.

    Parameters
    ----------
    date_array : array-like or pandas.DatetimeIndex
        Array-like or index of datetimes to be downloaded.
    data_path : str
        Path to data directory.
    local_file_prefix : str
        Prefix for local files, e.g., 'tag_' or 'tag_monthly_'
    local_date_fmt : str
        String format for the local filename, e.g., '%Y-%m-%d' or '%Y-%m'
    gfz_data_name : str
        Name of the data index on the GFZ server, expects one of: 'Kp', 'ap',
        'Ap', 'Cp', 'C9', 'Hp30', 'Hp60', 'ap30', 'ap60', 'SN', 'Fobs', 'Fadj'
        where SN is the international sunspot number and Fxxx is the observed
        and adjusted F10.7.
    freq : pds.DateOffset or dt.timedelta
        Offset to add to the start date to ensure all data is downloaded
        (inclusive)
    update_files : bool
        Re-download data for files that already exist if True (default=False)
    is_def : bool
        If true, selects only the definitive data, otherwise also includes
        nowcast data (default=False)
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date (following the local
        file prefix and date format) as if they were downloaded (default=None)

    Raises
    ------
    IOError
        If there is a gateway timeout when downloading data or if an unknown
        mock download directory is supplied.

    """

    # If a mock download directory was supplied, test to see it exists
    if mock_download_dir is not None:
        if not os.path.isdir(mock_download_dir):
            raise IOError('file location is not a directory: {:}'.format(
                mock_download_dir))

    # Set the local variables
    base_url = "https://kp.gfz-potsdam.de/app/json/"
    time_fmt = "%Y-%m-%dT%H:%M:%SZ"
    last_file = ''

    # Cycle through all the dates
    for dl_date in date_array:
        # Build the local filename
        local_fname = ''.join([local_file_prefix,
                               dl_date.strftime(local_date_fmt), '.txt'])
        local_file = os.path.join(data_path, local_fname)

        # Determine if the download should occur
        if not os.path.isfile(local_file) or (update_files
                                              and local_file != last_file):
            if mock_download_dir is None:
                # Get the URL for the desired data
                stop = dl_date + freq
                query_url = "{:s}?start={:s}&end={:s}&index={:s}".format(
                    base_url, dl_date.strftime(time_fmt),
                    stop.strftime(time_fmt), gfz_data_name)

                if is_def:
                    # Add the definitive flag
                    query_url = '{:s}&status=def'.format(query_url)

                # The data is returned as a JSON file
                req = requests.get(query_url)

                # Process the JSON file
                if req.text.find('Gateway Timeout') >= 0:
                    raise IOError(''.join(['Gateway timeout when requesting ',
                                           'file using command: ', query_url]))

                raw_txt = req.text if req.ok else None
            else:
                # Get the text from the downloaded file
                query_url = os.path.join(mock_download_dir, local_fname)
                if os.path.isfile(query_url):
                    with open(query_url, 'r') as fpin:
                        raw_txt = fpin.read()
                else:
                    raw_txt = None

            if raw_txt is not None:
                raw_dict = json.loads(raw_txt)
                data = pds.DataFrame.from_dict({gfz_data_name:
                                                raw_dict[gfz_data_name]})
                if data.empty:
                    pysat.logger.warning("no data for {:}".format(dl_date))
                else:
                    # Convert the datetime strings to datetime objects
                    time_list = [dt.datetime.strptime(time_str, time_fmt)
                                 for time_str in raw_dict['datetime']]

                    # Add the time index
                    data.index = time_list

                    # Create a local CSV file
                    data.to_csv(local_file, header=True)
            else:
                pysat.logger.info("".join(["Data not downloaded for ",
                                           dl_date.strftime("%d %b %Y"),
                                           ", date may be out of range ",
                                           "for the database or data may ",
                                           "have been saved to an unexpected",
                                           " filename: {:}".format(query_url)]))
    return


def kp_ap_cp_download(platform, name, date_array, tag, inst_id, data_path,
                      mock_download_dir=None):
    """Download Kp, ap, and Cp data from GFZ.

    Parameters
    ----------
    platform : str
        Instrument platform.
    name : str
        Instrument name.
    date_array : array-like or pandas.DatetimeIndex
        Array-like or index of datetimes to be downloaded.
    tag : str
        String specifying the database, expects 'def' (definitive) or 'now'
        (nowcast)
    inst_id : str
        Specifies the instrument identification, not used.
    data_path : str
        Path to data directory.
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date (following the local
        file prefix and date format) as if they were downloaded (default=None)

    Raises
    ------
    ValueError
        If an unknown instrument module is supplied.
    IOError
        If an unknown mock download directory is supplied.

    Note
    ----
    Note that the download path for the complementary Instrument will use
    the standard pysat data paths

    """
    # If a mock download directory was supplied, test to see it exists
    if mock_download_dir is not None:
        if not os.path.isdir(mock_download_dir):
            raise IOError('file location is not a directory: {:}'.format(
                mock_download_dir))

    # Set the page for the definitive or nowcast Kp
    burl = ''.join(['https://datapub.gfz-potsdam.de/download/10.5880.Kp.0001',
                    '/Kp_', 'nowcast' if tag == 'now' else 'definitive', '/'])
    data_cols = ['Bartels_solar_rotation_num', 'day_within_Bartels_rotation',
                 'Kp', 'daily_Kp_sum', 'ap', 'daily_Ap', 'Cp', 'C9']
    hours = np.arange(0, 24, 3)
    kp_translate = {'0': 0.0, '3': 1.0 / 3.0, '7': 2.0 / 3.0}
    dnames = list()

    inst_cols = {'sw_kp': [0, 1, 2, 3], 'sw_cp': [0, 1, 6, 7],
                 'sw_ap': [0, 1, 4, 5]}

    # Construct the Instrument module name from the platform and name
    inst_mod_name = '_'.join([platform, name])
    if inst_mod_name not in inst_cols.keys():
        raise ValueError('Unknown Instrument module {:}, expected {:}'.format(
            inst_mod_name, inst_cols.keys()))

    data_paths = {inst_mod: data_path if inst_mod == inst_mod_name else
                  general.get_instrument_data_path(inst_mod, tag=tag,
                                                   inst_id=inst_id)
                  for inst_mod in inst_cols.keys()}

    # Check that the directories exist
    for data_path in data_paths.values():
        pysat.utils.files.check_and_make_path(data_path)

    # Cycle through all the times
    for dl_date in date_array:
        fname = 'Kp_{:s}{:04d}.wdc'.format(tag, dl_date.year)
        if fname not in dnames:
            pysat.logger.info(' '.join(('Downloading file for',
                                        dl_date.strftime('%Y'))))
            if mock_download_dir is None:
                furl = ''.join([burl, fname])
                req = requests.get(furl)

                raw_txt = req.text if req.ok else None
            else:
                furl = os.path.join(mock_download_dir, fname)
                if os.path.isfile(furl):
                    with open(furl, "r") as fpin:
                        raw_txt = fpin.read()
                else:
                    raw_txt = None

            if raw_txt is not None:
                # Split the file text into lines
                lines = raw_txt.split('\n')[:-1]

                # Remove the header
                while lines[0].find('#') == 0:
                    lines.pop(0)

                # Process the data lines
                ddict = {dkey: list() for dkey in data_cols}
                times = list()
                for line in lines:
                    ldate = dt.datetime.strptime(' '.join([
                        "{:02d}".format(int(dd)) for dd in
                        [line[:2], line[2:4], line[4:6]]]), "%y %m %d")
                    bsr_num = int(line[6:10])
                    bsr_day = int(line[10:12])
                    if line[28:30] == '  ':
                        kp_ones = 0.0
                    else:
                        kp_ones = float(line[28:30])
                    sum_kp = kp_ones + kp_translate[line[30]]
                    daily_ap = int(line[55:58])
                    cp = float(line[58:61])
                    c9 = int(line[61])

                    for i, hour in enumerate(hours):
                        # Set the time for this hour and day
                        times.append(ldate + dt.timedelta(hours=int(hour)))

                        # Set the daily values for this hour
                        ddict['Bartels_solar_rotation_num'].append(bsr_num)
                        ddict['day_within_Bartels_rotation'].append(bsr_day)
                        ddict['daily_Kp_sum'].append(sum_kp)
                        ddict['daily_Ap'].append(daily_ap)
                        ddict['Cp'].append(cp)
                        ddict['C9'].append(c9)

                        # Get the hourly-specific values
                        ikp = i * 2
                        kp_ones = line[12 + ikp]
                        if kp_ones == ' ':
                            kp_ones = 0.0
                        ddict['Kp'].append(np.float64(kp_ones)
                                           + kp_translate[line[13 + ikp]])
                        iap = i * 3
                        ddict['ap'].append(np.int64(line[31 + iap:34 + iap]))

                # Put data into nicer DataFrames
                for inst_mod in inst_cols.keys():
                    sel_cols = np.array(data_cols)[inst_cols[inst_mod]]
                    sel_dict = {col: ddict[col] for col in sel_cols}
                    data = pds.DataFrame(sel_dict, index=times,
                                         columns=sel_cols)

                    # Write out as a CSV file
                    sfname = fname.replace('Kp',
                                           inst_mod.split('_')[-1].capitalize())
                    saved_fname = os.path.join(data_paths[inst_mod],
                                               sfname).replace('.wdc', '.txt')
                    data.to_csv(saved_fname, header=True)

                # Record the filename so we don't download it twice
                dnames.append(fname)
            else:
                pysat.logger.info("".join(["Unable to download data for ",
                                           dl_date.strftime("%d %b %Y"),
                                           ", date may be out of range for ",
                                           "the database or data may have been",
                                           " saved to an unexpected filename: ",
                                           furl]))
    return


def kp_ap_cp_list_files(name, tag, inst_id, data_path, format_str=None):
    """List local files for Kp, ap, or Cp data obtained from GFZ.

    Parameters
    ----------
    name : str
        Instrument name.
    tag : str
        String specifying the database, expects 'def' (definitive) or 'now'
        (nowcast)
    inst_id : str
        Specifies the instrument identification, not used.
    data_path : str
        Path to data directory.
    format_str : str or NoneType
        User specified file format.  If None is specified, the default
        formats associated with the supplied tags are used. (default=None)

    Returns
    -------
    files : pysat._files.Files
        A class containing the verified available files

    """

    if format_str is None:
        format_str = ''.join(['_'.join([name.capitalize(), tag]),
                              '{year:04d}.txt'])

    # Files are stored by year, going to add a date to the yearly filename for
    # each month and day of month.  The load routine will load the year and use
    # the append date to select out approriate data.
    files = pysat.Files.from_os(data_path=data_path, format_str=format_str)
    if not files.empty:
        files.loc[files.index[-1]
                  + pds.DateOffset(years=1, days=-1)] = files.iloc[-1]
        files = files.asfreq('D', 'pad')
        files = files + '_' + files.index.strftime('%Y-%m-%d')

    return files


def load_def_now(fnames):
    """Load GFZ yearly definitive or nowcast index data.

    Parameters
    ----------
    fnames : pandas.Series
        Series of filenames

    Returns
    -------
    data : pandas.DataFrame
        Object containing satellite data

    """

    # Load the definitive or nowcast data. The GFZ index data are stored in
    # yearly or monthly files that are separated by index ondownload. We need
    # to return data daily. The daily date is attached to filename. Parse off
    # the last date, load all data, and downselect to the desired day
    unique_fnames = dict()
    for filename in fnames:
        fname = filename[0:-11]
        fdate = dt.datetime.strptime(filename[-10:], '%Y-%m-%d')
        if fname not in unique_fnames.keys():
            unique_fnames[fname] = [fdate]
        else:
            unique_fnames[fname].append(fdate)

    # Load the desired filenames
    all_data = []
    for fname in unique_fnames.keys():
        # The daily date is attached to the filename.  Parse off the last
        # date, load the year of data, downselect to the desired day
        fdate = min(unique_fnames[fname])
        temp = pds.read_csv(fname, index_col=0, parse_dates=True)

        if temp.empty:
            pysat.logger.warn('Empty file: {:}'.format(fname))
            continue

        # Select the desired times and add to data list
        all_data.append(pds.DataFrame(temp[fdate:max(unique_fnames[fname])
                                           + dt.timedelta(seconds=86399)]))

    # Combine data together
    if len(all_data) > 0:
        data = pds.concat(all_data, axis=0, sort=True)
    else:
        data = pds.DataFrame()

    return data
